#!/usr/bin/env python

from leather.axis import Axis
from leather.chart import Chart
from leather.data_types import Date, DateTime
from leather.grid import Grid
from leather.scales import Scale, Linear
from leather.series import Series
from leather.shapes import Line
from leather import theme
from leather.utils import X, Y


class Lattice(object):
    """
    A grid of charts with synchronized shapes, scales, and axes.

    Lattice only supports graphing a single series of data.

    :param shape:
        An instance of :class:`.Shape` to use to render all series. Defaults
        to :class:`.Line` if not specified.
    """
    def __init__(self, shape=None):
        self._shape = shape or Line()
        self._series = []
        self._types = [None, None]
        self._scales = [None, None]
        self._axes = [None, None]

    def set_x_scale(self, scale):
        """
        Set the X :class:`.Scale` for this lattice.
        """
        self._scales[X] = scale

    def set_y_scale(self, scale):
        """
        See :meth:`.Lattice.set_x_scale`.
        """
        self._scales[Y] = scale

    def add_x_scale(self, domain_min, domain_max):
        """
        Create and add a :class:`.Scale`.

        If the provided domain values are :class:`date` or :class:`datetime`
        then a :class:`.Temporal` scale will be created, otherwise it will
        :class:`.Linear`.

        If you want to set a custom scale class use :meth:`.Lattice.set_x_scale`
        instead.
        """
        scale_type = Linear

        if isinstance(domain_min, Date.types) or isinstance(domain_min, DateTime.types):
            scale_type = Temporal

        self.set_x_scale(scale_type(domain_min, domain_max))

    def add_y_scale(self, domain_min, domain_max):
        """
        See :meth:`.Lattice.add_x_scale`.
        """
        scale_type = Linear

        if isinstance(domain_min, Date.types) or isinstance(domain_min, DateTime.types):
            scale_type = Temporal

        self.set_y_scale(scale_type(domain_min, domain_max))

    def set_x_axis(self, axis):
        """
        Set an :class:`.Axis` class for this lattice.
        """
        self._axes[X] = axis

    def set_y_axis(self, axis):
        """
        See :meth:`.Lattice.set_x_axis`.
        """
        self._axes[Y] = axis

    def add_x_axis(self, ticks=None, tick_formatter=None, name=None):
        """
        Create and add an X :class:`.Axis`.

        If you want to set a custom axis class use :meth:`.Lattice.set_x_axis`
        instead.
        """
        self._axes[X] = Axis(ticks=ticks, tick_formatter=tick_formatter, name=name)

    def add_y_axis(self, ticks=None, tick_formatter=None, name=None):
        """
        See :meth:`.Lattice.add_x_axis`.
        """
        self._axes[Y] = Axis(ticks=ticks, tick_formatter=tick_formatter, name=name)

    def add_one(self, data, x=None, y=None, title=None):
        """
        Add a data series to this lattice.

        :param data:
            A sequence of data suitable for constructing a :class:`.Series`,
            or a sequence of such objects.
        :param x:
            See :class:`.Series`.
        :param y:
            See :class:`.Series`.
        :param title:
            A title to render above this chart.
        """
        series = Series(data, x=x, y=y, name=title)

        for dimension in [X, Y]:
            if self._types[dimension]:
                if series._types[dimension] is not self._types[dimension]:
                    raise TypeError('All data series must have the same data types.')
            else:
                self._types[dimension] = series._types[dimension]

        self._shape.validate_series(series)
        self._series.append(series)

    def add_many(self, data, x=None, y=None, titles=None):
        """
        Same as :meth:`.Lattice.add_one` except :code:`data` is a list of data
        series to be added simultaneously.

        See :meth:`.Lattice.add_one` for other arguments.

        Note that :code:`titles` is a sequence of titles that must be the same
        length as :code:`data`.
        """
        for i, d in enumerate(data):
            title = titles[i] if titles else None

            self.add_one(d, x=x, y=y, title=title)

    def to_svg(self, path=None, width=None, height=None):
        """
        Render the lattice to an SVG.

        See :class:`.Grid` for additional documentation.
        """
        layers = [(s, self._shape) for s in self._series]

        if not self._scales[X]:
            self._scales[X] = Scale.infer(layers, X, self._types[X])

        if not self._scales[Y]:
            self._scales[Y]= Scale.infer(layers, Y, self._types[Y])

        if not self._axes[X]:
            self._axes[X] = Axis()

        if not self._axes[Y]:
            self._axes[Y] = Axis()

        grid = Grid()

        for i, series in enumerate(self._series):
            chart = Chart(title=series.name)
            chart.set_x_scale(self._scales[X])
            chart.set_y_scale(self._scales[Y])
            chart.set_x_axis(self._axes[X])
            chart.set_y_axis(self._axes[Y])
            chart.add_series(series, self._shape)

            grid.add_one(chart)

        return grid.to_svg(path, width, height)



#!/usr/bin/env python

import xml.etree.ElementTree as ET

import six

from leather.data_types import Number, Text
from leather.series import CategorySeries
from leather.shapes.base import Shape
from leather import theme
from leather.utils import X, Y


class Bars(Shape):
    """
    Render a series of data as bars.

    :param fill_color:
        The color to fill the bars. You may also specify a
        :func:`.style_function`.
    """
    def __init__(self, fill_color=None):
        self._fill_color = fill_color

    def validate_series(self, series):
        """
        Verify this shape can be used to render a given series.
        """
        if isinstance(series, CategorySeries):
            raise ValueError('Bars can not be used to render CategorySeries.')

        if series.data_type(X) is not Number:
            raise ValueError('Bars only support Number values for the Y axis.')

        if series.data_type(Y) is not Text:
            raise ValueError('Bars only support Text values for the X axis.')

    def to_svg(self, width, height, x_scale, y_scale, series, palette):
        """
        Render bars to SVG elements.
        """
        group = ET.Element('g')
        group.set('class', 'series bars')

        zero_x = x_scale.project(0, 0, width)

        if self._fill_color:
            fill_color = self._fill_color
        else:
            fill_color = next(palette)

        for d in series.data():
            if d.x is None or d.y is None:
                continue

            y1, y2 = y_scale.project_interval(d.y, height, 0)
            proj_x = x_scale.project(d.x, 0, width)

            if d.x < 0:
                bar_x = proj_x
                bar_width = zero_x - proj_x
            else:
                bar_x = zero_x
                bar_width = proj_x - zero_x

            if callable(fill_color):
                color = fill_color(d)
            else:
                color = fill_color

            group.append(ET.Element('rect',
                x=six.text_type(bar_x),
                y=six.text_type(y2),
                width=six.text_type(bar_width),
                height=six.text_type(y1 - y2),
                fill=color
            ))

        return group



#!/usr/bin/env python

from datetime import date, datetime

import six

from leather.data_types import Date, DateTime, Number, Text
from leather.shapes import Bars, Columns


class Scale(object):
    """
    Base class for various kinds of scale objects.
    """
    @classmethod
    def infer(cls, layers, dimension, data_type):
        """
        Infer's an appropriate default scale for a given sequence of
        :class:`.Series`.

        :param chart_series:
            A sequence of :class:`.Series` instances
        :param dimension:
            The dimension, :code:`X` or :code:`Y` of the data to infer for.
        :param data_type:
            The type of data contained in the series dimension.
        """
        from leather.scales.linear import Linear
        from leather.scales.ordinal import Ordinal
        from leather.scales.temporal import Temporal

        # Default Time scale is Temporal
        if data_type is Date:
            data_min = date.max
            data_max = date.min

            for series, shape in layers:
                data_min = min(data_min, series.min(dimension))
                data_max = max(data_max, series.max(dimension))

            scale = Temporal(data_min, data_max)
        elif data_type is DateTime:
            data_min = datetime.max
            data_max = datetime.min

            for series, shape in layers:
                data_min = min(data_min, series.min(dimension))
                data_max = max(data_max, series.max(dimension))

            scale = Temporal(data_min, data_max)
        # Default Number scale is Linear
        elif data_type is Number:
            force_zero = False
            data_min = None
            data_max = None

            for series, shape in layers:
                if isinstance(shape, (Bars, Columns)):
                    force_zero = True

                if data_min is None:
                    data_min = series.min(dimension)
                else:
                    data_min = min(data_min, series.min(dimension))

                if data_max is None:
                    data_max = series.max(dimension)
                else:
                    data_max = max(data_max, series.max(dimension))

            if force_zero:
                if data_min > 0:
                    data_min = 0

                if data_max < 0:
                    data_max = 0

            scale = Linear(data_min, data_max)
        # Default Text scale is Ordinal
        elif data_type is Text:
            scale_values = None

            # First case: a single set of ordinal labels
            if len(layers) == 1:
                scale_values = layers[0][0].values(dimension)
            else:
                first_series = set(layers[0][0].values(dimension))
                data_series = [series.values(dimension) for series, shape in layers]
                all_same = True

                for series in data_series:
                    if set(series) != first_series:
                        all_same = False
                        break

                # Second case: multiple identical sets of ordinal labels
                if all_same:
                    scale_values = layers[0][0].values(dimension)
                # Third case: multiple different sets of ordinal labels
                else:
                    scale_values = sorted(list(set().union(*data_series)))

            scale = Ordinal(scale_values)

        return scale

    def contains(self, v):
        """
        Return :code:`True` if a given value is contained within this scale's
        displayed domain.
        """
        raise NotImplementedError

    def project(self, value, range_min, range_max):
        """
        Project a value in this scale's domain to a target range.
        """
        raise NotImplementedError

    def project_interval(self, value, range_min, range_max):
        """
        Project a value in this scale's domain to an interval in the target
        range. This is used for places :class:`.Bars` and :class:`.Columns`.
        """
        raise NotImplementedError

    def ticks(self):
        """
        Generate a series of ticks for this scale.
        """
        raise NotImplementedError

    def format_tick(self, value, i, count):
        """
        Format ticks for display.

        This method is used as a default which will be ignored if the user
        provides a custom tick formatter to the axis.
        """
        return six.text_type(value)
